/*********************************************************************
 *
 * This is based on code created by Peter Harvey,
 * (pharvey@codebydesign.com).
 *
 * Modified and extended by Nick Gorham
 * (nick@easysoft.com).
 *
 * Any bugs or problems should be considered the fault of Nick and not
 * Peter.
 *
 * copyright (c) 1999 Nick Gorham
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **********************************************************************
 *
 * $Id: SQLError.c,v 1.7 1999/08/03 21:47:39 shandyb Exp $
 *
 * $Log: SQLError.c,v $
 * Revision 1.7  1999/08/03 21:47:39  shandyb
 * Moving to automake: changed files in DriverManager
 *
 * Revision 1.6  1999/07/15 06:22:33  ngorham
 *
 * Fixed spelling mistake
 *
 * Revision 1.5  1999/07/14 19:46:04  ngorham
 *
 * Fix the error logging when SQLError or SQLGetDiagRec returns SQL_NO_DATA
 *
 * Revision 1.4  1999/07/12 19:42:05  ngorham
 *
 * Finished off SQLGetDiagField.c and fixed a but that caused SQLError to
 * fail with Perl and PHP, connect errors were not being returned because
 * I was checking to the environment being set, they were setting the
 * statement and the environment. The order of checking has been changed.
 *
 * Revision 1.3  1999/07/04 21:05:07  ngorham
 *
 * Add LGPL Headers to code
 *
 * Revision 1.2  1999/06/30 23:56:54  ngorham
 *
 * Add initial thread safety code
 *
 * Revision 1.1.1.1  1999/05/29 13:41:06  sShandyb
 * first go at it
 *
 * Revision 1.3  1999/06/02 20:12:10  ngorham
 *
 * Fixed botched log entry, and removed the dos \r from the sql header files.
 *
 * Revision 1.2  1999/06/02 19:57:20  ngorham
 *
 * Added code to check if a attempt is being made to compile with a C++
 * Compiler, and issue a message.
 * Start work on the ODBC2-3 conversions.
 *
 * Revision 1.1.1.1  1999/05/27 18:23:17  pharvey
 * Imported sources
 *
 * Revision 1.3  1999/04/30 16:22:47  nick
 * Another checkpoint
 *
 * Revision 1.2  1999/04/29 21:40:58  nick
 * End of another night :-)
 *
 * Revision 1.1  1999/04/25 23:06:11  nick
 * Initial revision
 *
 *
 **********************************************************************/

#include "drivermanager.h"

static char const rcsid[]= "$RCSfile: SQLError.c,v $ $Revision: 1.7 $";

SQLRETURN SQLError( SQLHENV environment_handle,
           SQLHDBC connection_handle,
           SQLHSTMT statement_handle,
           SQLCHAR *sqlstate,
           SQLINTEGER *native_error,
           SQLCHAR *message_text,
           SQLSMALLINT buffer_length,
           SQLSMALLINT *text_length )
{
    SQLRETURN ret;
    SQLCHAR s0[ 20 ], s1[ 100 ];

    if ( statement_handle )
    {
    DMHSTMT statement = (DMHSTMT) statement_handle;

        /*
         * check connection
         */

        if ( !__validate_stmt( statement ))
        {
            return SQL_INVALID_HANDLE;
        }

        if ( sqlstate )
            strcpy( sqlstate, "00000" );

        sprintf( statement -> msg, 
            "\n\t\tEntry:\
            \n\t\t\tStatement = %p\
            \n\t\t\tSQLState = %p\
            \n\t\t\tNative = %p\
            \n\t\t\tMessage Text = %p\
            \n\t\t\tBuffer Length = %d\
            \n\t\t\tText Len Ptr = %p",
                statement,
                sqlstate,
                native_error,
                message_text,
                buffer_length,
                text_length );

        logPushMsg( statement -> log_handle, 
                __get_pid( s0 ), 
                __FILE__, 
                __LINE__, 
                LOG_INFO, 
                LOG_INFO, 
                statement -> msg );

        thread_protect( SQL_HANDLE_STMT, statement );

        /*
         * get any local ones
         */
        if ( statement -> error.internal_count > 0 )
        {
            char msg[ SQL_MAX_MESSAGE_LENGTH + 40 ];

            sprintf( msg, "[unixODBC Driver Manager]%s",
                    statement -> error.msg );

            ret = SQL_SUCCESS;

            if ( sqlstate )
                strcpy( sqlstate, statement -> error.sqlstate );
            if ( buffer_length + 1 < strlen( msg ))
            {
                if ( message_text )
                {
                    memcpy( message_text,
                            msg,
                            buffer_length - 1 );
                    message_text[ buffer_length - 1 ] = '\0';
                }
                ret = SQL_SUCCESS_WITH_INFO;
            }
            else
            {
                if ( message_text )
                    strcpy( message_text, msg );
            }
            if ( text_length )
                *text_length = strlen( msg );

            if ( native_error )
                *native_error = statement -> error.native_error;

            statement -> error.internal_count --;
        }
        else if ( !CHECK_SQLERROR( statement -> connection ))
        {
            sprintf( statement -> msg, 
                "\n\t\tExit:[%s]",
                    __get_return_status( SQL_NO_DATA ));

            logPushMsg( statement -> log_handle, 
                    __get_pid( s0 ), 
                    __FILE__, 
                    __LINE__, 
                    LOG_INFO, 
                    LOG_INFO, 
                    statement -> msg );

            thread_release( SQL_HANDLE_STMT, statement );

            return SQL_NO_DATA;
        }
        else
        {
            ret = SQLERROR( statement -> connection,
                    SQL_NULL_HENV,
                    SQL_NULL_HDBC,
                    statement -> driver_stmt,
                    sqlstate,
                    native_error,
                    message_text,
                    buffer_length,
                    text_length );
        }

        if ( SQL_SUCCEEDED( ret ))
        {
            sprintf( statement -> msg, 
                "\n\t\tExit:[%s]\
                \n\t\t\tSQLState = %s\
                \n\t\t\tNative = %s\
                \n\t\t\tMessage Text = %s",
                    __get_return_status( ret ),
                    sqlstate,
                    __ptr_as_string( s0, native_error ),
                    __sdata_as_string( s1, SQL_CHAR, 
                        text_length, message_text ));
        }
        else
        {
            sprintf( statement -> msg, 
                "\n\t\tExit:[%s]",
                    __get_return_status( ret ));
        }

        logPushMsg( statement -> log_handle, 
                __get_pid( s0 ), 
                __FILE__, 
                __LINE__, 
                LOG_INFO, 
                LOG_INFO, 
                statement -> msg );

        thread_release( SQL_HANDLE_STMT, statement );

        return ret;
    }
    else if ( connection_handle )
    {
    DMHDBC connection = (DMHDBC) connection_handle;

        /*
         * check connection
         */

        if ( !__validate_dbc( connection ))
        {
            return SQL_INVALID_HANDLE;
        }

        if ( sqlstate )
            strcpy( sqlstate, "00000" );

        sprintf( connection -> msg, 
            "\n\t\tEntry:\
            \n\t\t\tConnection = %p\
            \n\t\t\tSQLState = %p\
            \n\t\t\tNative = %p\
            \n\t\t\tMessage Text = %p\
            \n\t\t\tBuffer Length = %d\
            \n\t\t\tText Len Ptr = %p",
                connection,
                sqlstate,
                native_error,
                message_text,
                buffer_length,
                text_length );

        logPushMsg( connection -> log_handle, 
                __get_pid( s0 ), 
                __FILE__, 
                __LINE__, 
                LOG_INFO, 
                LOG_INFO, 
                connection -> msg );

        thread_protect( SQL_HANDLE_DBC, connection );

        /*
         * get any local ones
         */
        if ( connection -> error.internal_count > 0 )
        {
            char msg[ SQL_MAX_MESSAGE_LENGTH + 40 ];

            sprintf( msg, "[unixODBC Driver Manager]%s",
                    connection -> error.msg );

            ret = SQL_SUCCESS;

            if ( sqlstate )
            strcpy( sqlstate, connection -> error.sqlstate );
            if ( buffer_length + 1 < strlen( msg ))
            {
                if ( message_text )
                {
                    memcpy( message_text,
                            msg,
                            buffer_length - 1 );
                    message_text[ buffer_length - 1 ] = '\0';
                }
                ret = SQL_SUCCESS_WITH_INFO;
            }
            else
            {
                if ( message_text )
                    strcpy( message_text, msg );
            }
            if ( text_length )
                *text_length = strlen( msg );

            if ( native_error )
                *native_error = connection -> error.native_error;

            connection -> error.internal_count --;
        }
        else if ( connection -> state < STATE_C4 )
        {
            ret = SQL_NO_DATA;
        }
        else if ( CHECK_SQLERROR( connection ))
        {
            ret = SQLERROR( connection,
                    SQL_NULL_HENV,
                    connection -> driver_dbc,
                    SQL_NULL_HSTMT,
                    sqlstate,
                    native_error,
                    message_text,
                    buffer_length,
                    text_length );
        }
        else
        {
            sprintf( connection -> msg, 
                "\n\t\tExit:[%s]",
                    __get_return_status( SQL_NO_DATA ));

            logPushMsg( connection -> log_handle, 
                    __get_pid( s0 ), 
                    __FILE__, 
                    __LINE__, 
                    LOG_INFO, 
                    LOG_INFO, 
                    connection -> msg );

            thread_release( SQL_HANDLE_DBC, connection );

            return SQL_NO_DATA;
        }

        if ( SQL_SUCCEEDED( ret ))
        {
            sprintf( connection -> msg, 
                "\n\t\tExit:[%s]\
                \n\t\t\tSQLState = %s\
                \n\t\t\tNative = %s\
                \n\t\t\tMessage Text = %s",
                    __get_return_status( ret ),
                    sqlstate,
                    __ptr_as_string( s0, native_error ),
                    __sdata_as_string( s1, SQL_CHAR, 
                        text_length, message_text ));
        }
        else
        {
            sprintf( connection -> msg, 
                "\n\t\tExit:[%s]",
                    __get_return_status( ret ));
        }

        logPushMsg( connection -> log_handle, 
                __get_pid( s0 ), 
                __FILE__, 
                __LINE__, 
                LOG_INFO, 
                LOG_INFO, 
                connection -> msg );

        thread_release( SQL_HANDLE_DBC, connection );

        return ret;
    }
    else if ( environment_handle )
    {
    DMHENV environment = (DMHENV) environment_handle;

        /*
         * check environment
         */

        if ( !__validate_env( environment ))
        {
            return SQL_INVALID_HANDLE;
        }

        if ( sqlstate )
            strcpy( sqlstate, "00000" );

        sprintf( environment -> msg, 
            "\n\t\tEntry:\
            \n\t\t\tEnvironment = %p\
            \n\t\t\tSQLState = %p\
            \n\t\t\tNative = %p\
            \n\t\t\tMessage Text = %p\
            \n\t\t\tBuffer Length = %d\
            \n\t\t\tText Len Ptr = %p",
                environment,
                sqlstate,
                native_error,
                message_text,
                buffer_length,
                text_length );

        logPushMsg( environment -> log_handle, 
                __get_pid( s0 ), 
                __FILE__, 
                __LINE__, 
                LOG_INFO, 
                LOG_INFO, 
                environment -> msg );

        thread_protect( SQL_HANDLE_ENV, environment );

        /*
         * we can only look for local ones
         * unless the DM starts polling the driver after calls
         */
        if ( environment -> error.internal_count > 0 )
        {
            char msg[ SQL_MAX_MESSAGE_LENGTH + 40 ];

            sprintf( msg, "[unixODBC Driver Manager]%s",
                    environment -> error.msg );

            if ( sqlstate )
                strcpy( sqlstate, environment -> error.sqlstate );
            if ( buffer_length < strlen( msg ))
            {
                if ( message_text )
                {
                    memcpy( message_text,
                            msg,
                            buffer_length - 1 );
                    message_text[ buffer_length - 1 ] = '\0';
                }
                ret = SQL_SUCCESS_WITH_INFO;
            }
            else
            {
                if ( message_text )
                    strcpy( message_text, msg );
                 ret = SQL_SUCCESS;
            }
            if ( text_length )
                *text_length = strlen( msg );

            if ( native_error )
                *native_error = environment -> error.native_error;

            environment -> error.internal_count --;

            sprintf( environment -> msg, 
                    "\n\t\tExit:[%s]\
                    \n\t\t\tSQLState = %s\
                    \n\t\t\tNative = %s\
                    \n\t\t\tMessage Text = %s",
                        __get_return_status( ret ),
                        sqlstate,
                        __ptr_as_string( s0, native_error ),
                        __sdata_as_string( s1, SQL_CHAR, 
                            text_length, message_text ));

            logPushMsg( environment -> log_handle, 
                    __get_pid( s0 ), 
                    __FILE__, 
                    __LINE__, 
                    LOG_INFO, 
                    LOG_INFO, 
                    environment -> msg );

            thread_release( SQL_HANDLE_ENV, environment );

            return ret;
        }
        else
        {
            sprintf( environment -> msg, 
                    "\n\t\tExit:[%s]",
                        __get_return_status( SQL_NO_DATA ));

            logPushMsg( environment -> log_handle, 
                    __get_pid( s0 ), 
                    __FILE__, 
                    __LINE__, 
                    LOG_INFO, 
                    LOG_INFO, 
                    environment -> msg );

            thread_release( SQL_HANDLE_ENV, environment );

            return SQL_NO_DATA;
        }
    }
    else
    {
        return SQL_INVALID_HANDLE;
    }
}
