/*********************************************************************
 *
 * This is based on code created by Peter Harvey,
 * (pharvey@codebydesign.com).
 *
 * Modified and extended by Nick Gorham
 * (nick@easysoft.com).
 *
 * Any bugs or problems should be considered the fault of Nick and not
 * Peter.
 *
 * copyright (c) 1999 Nick Gorham
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **********************************************************************
 *
 * $Id: SQLGetDiagField.c,v 1.4 1999/07/12 19:42:06 ngorham Exp $
 *
 * $Log: SQLGetDiagField.c,v $
 * Revision 1.4  1999/07/12 19:42:06  ngorham
 *
 * Finished off SQLGetDiagField.c and fixed a but that caused SQLError to
 * fail with Perl and PHP, connect errors were not being returned because
 * I was checking to the environment being set, they were setting the
 * statement and the environment. The order of checking has been changed.
 *
 * Revision 1.3  1999/07/04 21:05:07  ngorham
 *
 * Add LGPL Headers to code
 *
 * Revision 1.2  1999/06/30 23:56:55  ngorham
 *
 * Add initial thread safety code
 *
 * Revision 1.1.1.1  1999/05/29 13:41:07  sShandyb
 * first go at it
 *
 * Revision 1.1.1.1  1999/05/27 18:23:17  pharvey
 * Imported sources
 *
 * Revision 1.5  1999/05/09 23:27:11  nick
 * All the API done now
 *
 * Revision 1.4  1999/05/04 22:41:12  nick
 * and another night ends
 *
 * Revision 1.3  1999/05/03 19:50:43  nick
 * Another check point
 *
 * Revision 1.2  1999/04/30 16:22:47  nick
 * Another checkpoint
 *
 * Revision 1.1  1999/04/25 23:06:11  nick
 * Initial revision
 *
 *
 **********************************************************************/

#include "drivermanager.h"

static char const rcsid[]= "$RCSfile: SQLGetDiagField.c,v $ $Revision: 1.4 $";

SQLRETURN SQLGetDiagField( SQLSMALLINT handle_type,
        SQLHANDLE handle,
        SQLSMALLINT rec_number,
        SQLSMALLINT diag_identifier,
        SQLPOINTER diag_info_ptr,
        SQLSMALLINT buffer_length,
        SQLSMALLINT *string_length_ptr )
{
    SQLCHAR s0[ 20 ], s1[ 100 ];

    switch ( handle_type )
    {
      case SQL_HANDLE_ENV:
        {
            DMHENV environment = (DMHENV) handle;

            /*
             * check environment
             */

            if ( !__validate_env( environment ))
            {
                return SQL_INVALID_HANDLE;
            }

            sprintf( environment -> msg, "\n\t\tEntry:\
                \n\t\t\tEnvironment = %p\
                \n\t\t\tRec Number = %d\
                \n\t\t\tDiag Ident = %s\
                \n\t\t\tBuffer Len = %d\
                \n\t\t\tStrLen = %p",
                    environment,
                    rec_number,
                    __diag_attr_as_string( s1, diag_identifier ),
                    (int)buffer_length,
                    (void*)string_length_ptr );

            logPushMsg( environment -> log_handle, 
                    __get_pid( s0 ), 
                    __FILE__, 
                    __LINE__, 
                    LOG_INFO, 
                    LOG_INFO, 
                    environment -> msg );

            thread_protect( SQL_HANDLE_ENV, environment );

            if ( diag_identifier == SQL_DIAG_NUMBER )
            {
                if ( diag_info_ptr )
                {
                    *((SQLINTEGER*) diag_info_ptr ) =
                        environment -> error.internal_count;
                }

                sprintf( environment -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( SQL_SUCCESS ));

                logPushMsg( environment -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        environment -> msg );

                thread_release( SQL_HANDLE_ENV, environment );

                return SQL_SUCCESS;
            }
        }
        break;

      case SQL_HANDLE_DBC:
        {
            DMHDBC connection = (DMHDBC) handle;
            int count = 0;
            SQLRETURN ret;

            /*
             * check connection
             */

            if ( !__validate_dbc( connection ))
            {
                return SQL_INVALID_HANDLE;
            }

            sprintf( connection -> msg, "\n\t\tEntry:\
                \n\t\t\tConnection = %p\
                \n\t\t\tRec Number = %d\
                \n\t\t\tDiag Ident = %s\
                \n\t\t\tBuffer Len = %d\
                \n\t\t\tStrLen = %p",
                    connection,
                    rec_number,
                    __diag_attr_as_string( s1, diag_identifier ),
                    (int)buffer_length,
                    (void*)string_length_ptr );

            logPushMsg( connection -> log_handle, 
                    __get_pid( s0 ), 
                    __FILE__, 
                    __LINE__, 
                    LOG_INFO, 
                    LOG_INFO, 
                    connection -> msg );

            thread_protect( SQL_HANDLE_DBC, connection );

            count = connection -> error.internal_count;

            if ( connection -> state >= STATE_C4 &&
                    CHECK_SQLGETDIAGFIELD( connection ))
            {
                int cnt;
                ret = SQLGETDIAGFIELD( connection,
                                    handle_type,
                                    connection -> driver_dbc,
                                    0,
                                    SQL_DIAG_NUMBER,
                                    &cnt,
                                    sizeof( cnt ),
                                    NULL );

                if ( SQL_SUCCEEDED( ret ))
                {
                    count += cnt;
                }
            }

            if ( diag_identifier == SQL_DIAG_NUMBER )
            {
                if ( diag_info_ptr )
                {
                    *((SQLINTEGER*) diag_info_ptr ) = count;
                }
                sprintf( connection -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( SQL_SUCCESS ));

                logPushMsg( connection -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        connection -> msg );

                thread_release( SQL_HANDLE_DBC, connection );

                return SQL_SUCCESS;
            }
            else if ( rec_number < 1 )
            {
                sprintf( connection -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( SQL_ERROR ));

                logPushMsg( connection -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        connection -> msg );

                thread_release( SQL_HANDLE_DBC, connection );

                return SQL_SUCCESS;
            }
            else if ( rec_number <= connection -> error.internal_count )
            {
                int itype;
                int ival = 0;
                char *sval = "";
                char msg[ SQL_MAX_MESSAGE_LENGTH + 40 ];

                /*
                 * error from the DM
                 */
                switch ( diag_identifier )
                {
                  case SQL_DIAG_CURSOR_ROW_COUNT:
                  case SQL_DIAG_DYNAMIC_FUNCTION_CODE:
                  case SQL_DIAG_ROW_COUNT:
                  case SQL_DIAG_COLUMN_NUMBER:
                  case SQL_DIAG_NATIVE:
                  case SQL_DIAG_ROW_NUMBER:
                    itype = 1;
                    break;

                  case SQL_DIAG_RETURNCODE:
                    itype = 1;
                    ival = SQL_ERROR;
                    break;

                  case SQL_DIAG_DYNAMIC_FUNCTION:
                  case SQL_DIAG_CONNECTION_NAME:
                  case SQL_DIAG_SERVER_NAME:
                  case SQL_DIAG_SUBCLASS_ORIGIN:
                    itype = 2;
                    break;

                  case SQL_DIAG_MESSAGE_TEXT:
                    itype = 2;

                    sprintf( msg, "[unixODBC Driver Manager]%s",
                         connection -> error.msg );

                    sval = msg;
                    break;

                  case SQL_DIAG_SQLSTATE:
                    itype = 2;
                    sval = connection -> error.sqlstate;
                    break;

                  case SQL_DIAG_CLASS_ORIGIN:
                    itype = 2;
                    if ( strncmp( connection -> error.sqlstate, "IM", 2 ) == 0 )
                    {
                        sval = "ODBC 3.0";
                    }
                    else
                    {
                        sval = "ISO 9075";
                    }
                    break;
                }

                switch( itype )
                {
                  case 1:
                    if ( diag_info_ptr )
                        memcpy( diag_info_ptr, &ival, sizeof( ival ));

                    ret = SQL_SUCCESS;
                    break;

                  case 2:
                    if ( diag_info_ptr )
                    {
                        if ( buffer_length < strlen( sval ) + 1 )
                        {
                            ret = SQL_SUCCESS_WITH_INFO;
                            memcpy( diag_info_ptr, sval, buffer_length );
                            ((char*)diag_info_ptr)[ buffer_length - 1 ] = '\0';
                        }
                        else
                        {
                            ret = SQL_SUCCESS;
                            strcpy( diag_info_ptr, sval );
                        }
                    }
                    if ( string_length_ptr )
                    {
                        *string_length_ptr = strlen( sval ) + 1;
                    }
                    break;

                  default:
                    ret = SQL_ERROR;
                }

                sprintf( connection -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( ret ));

                logPushMsg( connection -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        connection -> msg );

                thread_release( SQL_HANDLE_DBC, connection );

                return ret;
            }
            else if ( CHECK_SQLGETDIAGFIELD( connection ))
            {
                /*
                 * error from the driver
                 */
                ret = SQLGETDIAGFIELD( connection,
                                    handle_type,
                                    connection -> driver_dbc,
                                    rec_number - 
                                        connection -> error.internal_count,
                                    diag_identifier,
                                    diag_info_ptr,
                                    buffer_length,
                                    string_length_ptr );

                sprintf( connection -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( SQL_ERROR ));

                logPushMsg( connection -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        connection -> msg );

                thread_release( SQL_HANDLE_DBC, connection );

                return ret;
            }
            else
            {
                sprintf( connection -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( SQL_ERROR ));

                logPushMsg( connection -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        connection -> msg );

                thread_release( SQL_HANDLE_DBC, connection );

                return SQL_ERROR;
            }
        }

      case SQL_HANDLE_STMT:
        {
            DMHSTMT statement  = (DMHSTMT) handle;
            int count;
            SQLRETURN ret;

            /*
             * check statement
             */

            if ( !__validate_stmt( statement ))
            {
                return SQL_INVALID_HANDLE;
            }

            sprintf( statement -> msg, "\n\t\tEntry:\
                \n\t\t\tStatement = %p\
                \n\t\t\tRec Number = %d\
                \n\t\t\tDiag Ident = %s\
                \n\t\t\tBuffer Len = %d\
                \n\t\t\tStrLen = %p",
                    statement,
                    rec_number,
                    __diag_attr_as_string( s1, diag_identifier ),
                    (int)buffer_length,
                    (void*)string_length_ptr );

            logPushMsg( statement -> log_handle, 
                    __get_pid( s0 ), 
                    __FILE__, 
                    __LINE__, 
                    LOG_INFO, 
                    LOG_INFO, 
                    statement -> msg );

            thread_protect( SQL_HANDLE_STMT, statement );

            count = statement -> error.internal_count;

            if ( CHECK_SQLGETDIAGFIELD( statement -> connection ))
            {
                int cnt;
                ret = SQLGETDIAGFIELD( statement -> connection,
                                    handle_type,
                                    statement -> driver_stmt,
                                    0,
                                    SQL_DIAG_NUMBER,
                                    &cnt,
                                    sizeof( cnt ),
                                    NULL );

                if ( SQL_SUCCEEDED( ret ))
                {
                    count += cnt;
                }
            }

            if ( diag_identifier == SQL_DIAG_NUMBER )
            {
                if ( diag_info_ptr )
                {
                    *((SQLINTEGER*) diag_info_ptr ) = count;
                }
                sprintf( statement -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( SQL_SUCCESS ));

                logPushMsg( statement -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        statement -> msg );

                thread_release( SQL_HANDLE_STMT, statement );

                return SQL_SUCCESS;
            }
            else if ( rec_number < 1 )
            {
                sprintf( statement -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( SQL_ERROR ));

                logPushMsg( statement -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        statement -> msg );

                thread_release( SQL_HANDLE_STMT, statement );

                return SQL_SUCCESS;
            }
            else if ( rec_number <= statement -> error.internal_count )
            {
                int itype;
                int ival = 0;
                char *sval = "";
                char msg[ SQL_MAX_MESSAGE_LENGTH + 40 ];

                /*
                 * error from the DM
                 */
                switch ( diag_identifier )
                {
                  case SQL_DIAG_CURSOR_ROW_COUNT:
                  case SQL_DIAG_DYNAMIC_FUNCTION_CODE:
                  case SQL_DIAG_ROW_COUNT:
                  case SQL_DIAG_COLUMN_NUMBER:
                  case SQL_DIAG_NATIVE:
                  case SQL_DIAG_ROW_NUMBER:
                    itype = 1;
                    break;

                  case SQL_DIAG_RETURNCODE:
                    itype = 1;
                    ival = SQL_ERROR;
                    break;

                  case SQL_DIAG_DYNAMIC_FUNCTION:
                  case SQL_DIAG_CONNECTION_NAME:
                  case SQL_DIAG_SERVER_NAME:
                  case SQL_DIAG_SUBCLASS_ORIGIN:
                    itype = 2;
                    break;

                  case SQL_DIAG_MESSAGE_TEXT:
                    itype = 2;

                    sprintf( msg, "[unixODBC Driver Manager]%s",
                         statement -> error.msg );

                    sval = msg;
                    break;

                  case SQL_DIAG_SQLSTATE:
                    itype = 2;
                    sval = statement -> error.sqlstate;
                    break;

                  case SQL_DIAG_CLASS_ORIGIN:
                    itype = 2;
                    if ( strncmp( statement -> error.sqlstate, "IM", 2 ) == 0 )
                    {
                        sval = "ODBC 3.0";
                    }
                    else
                    {
                        sval = "ISO 9075";
                    }
                    break;
                }

                switch( itype )
                {
                  case 1:
                    if ( diag_info_ptr )
                        memcpy( diag_info_ptr, &ival, sizeof( ival ));

                    ret = SQL_SUCCESS;
                    break;

                  case 2:
                    if ( diag_info_ptr )
                    {
                        if ( buffer_length < strlen( sval ) + 1 )
                        {
                            ret = SQL_SUCCESS_WITH_INFO;
                            memcpy( diag_info_ptr, sval, buffer_length );
                            ((char*)diag_info_ptr)[ buffer_length - 1 ] = '\0';
                        }
                        else
                        {
                            ret = SQL_SUCCESS;
                            strcpy( diag_info_ptr, sval );
                        }
                    }
                    if ( string_length_ptr )
                    {
                        *string_length_ptr = strlen( sval ) + 1;
                    }
                    break;

                  default:
                    ret = SQL_ERROR;
                }

                sprintf( statement -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( ret ));

                logPushMsg( statement -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        statement -> msg );

                thread_release( SQL_HANDLE_STMT, statement );

                return ret;
            }
            else if ( CHECK_SQLGETDIAGFIELD( statement -> connection ))
            {
                /*
                 * error from the driver
                 */
                ret = SQLGETDIAGFIELD( statement -> connection,
                                    handle_type,
                                    statement -> driver_stmt,
                                    rec_number - 
                                        statement -> error.internal_count,
                                    diag_identifier,
                                    diag_info_ptr,
                                    buffer_length,
                                    string_length_ptr );

                sprintf( statement -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( SQL_ERROR ));

                logPushMsg( statement -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        statement -> msg );

                thread_release( SQL_HANDLE_STMT, statement );

                return ret;
            }
            else
            {
                sprintf( statement -> msg, 
                        "\n\t\tExit:[%s]",
                        __get_return_status( SQL_ERROR ));

                logPushMsg( statement -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        statement -> msg );

                thread_release( SQL_HANDLE_STMT, statement );

                return SQL_ERROR;
            }
        }

      case SQL_HANDLE_DESC:
        return SQL_ERROR;

      default:
        return SQL_INVALID_HANDLE;
    }
    return SQL_NO_DATA;
}
