/*********************************************************************
 *
 * This is based on code created by Peter Harvey,
 * (pharvey@codebydesign.com).
 *
 * Modified and extended by Nick Gorham
 * (nick@easysoft.com).
 *
 * Any bugs or problems should be considered the fault of Nick and not
 * Peter.
 *
 * copyright (c) 1999 Nick Gorham
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **********************************************************************
 *
 * $Id: SQLGetDiagRec.c,v 1.4 1999/07/14 19:46:04 ngorham Exp $
 *
 * $Log: SQLGetDiagRec.c,v $
 * Revision 1.4  1999/07/14 19:46:04  ngorham
 *
 * Fix the error logging when SQLError or SQLGetDiagRec returns SQL_NO_DATA
 *
 * Revision 1.3  1999/07/04 21:05:07  ngorham
 *
 * Add LGPL Headers to code
 *
 * Revision 1.2  1999/06/30 23:56:55  ngorham
 *
 * Add initial thread safety code
 *
 * Revision 1.1.1.1  1999/05/29 13:41:07  sShandyb
 * first go at it
 *
 * Revision 1.1.1.1  1999/05/27 18:23:17  pharvey
 * Imported sources
 *
 * Revision 1.1  1999/04/30 16:22:47  nick
 * Another checkpoint
 *
 *
 **********************************************************************/

#include "drivermanager.h"

static char const rcsid[]= "$RCSfile: SQLGetDiagRec.c,v $ $Revision: 1.4 $";

SQLRETURN SQLGetDiagRec( SQLSMALLINT handle_type,
        SQLHANDLE   handle,
        SQLSMALLINT rec_number,
        SQLCHAR     *sqlstate,
        SQLINTEGER  *native,
        SQLCHAR     *message_text,
        SQLSMALLINT buffer_length,
        SQLSMALLINT *text_length_ptr )
{
    SQLRETURN ret;
    SQLCHAR s0[ 20 ], s1[ 100 ];

    if ( rec_number <= 0 )
    {
        return SQL_ERROR;
    }

    if ( buffer_length < 0 )
    {
        return SQL_ERROR;
    }

    switch ( handle_type )
    {
      case SQL_HANDLE_ENV:
        {
            DMHENV environment = (DMHENV) handle;

            /*
             * check environment
             */

            if ( !__validate_env( environment ))
            {
                return SQL_INVALID_HANDLE;
            }

            sprintf( environment -> msg, "\n\t\tEntry:\
                \n\t\t\tEnvironment = %p\
                \n\t\t\tRec Number = %d\
                \n\t\t\tSQLState = %p\
                \n\t\t\tNative = %p\
                \n\t\t\tMessage Text = %p\
                \n\t\t\tBuffer Length = %d\
                \n\t\t\tText Len Ptr = %p",
                    environment,
                    rec_number,
                    sqlstate,
                    native,
                    message_text,
                    buffer_length,
                    text_length_ptr );

            logPushMsg( environment -> log_handle, 
                    __get_pid( s0 ), 
                    __FILE__, 
                    __LINE__, 
                    LOG_INFO, 
                    LOG_INFO, 
                    environment -> msg );

            thread_protect( SQL_HANDLE_ENV, environment );

            /*
             * we can only look for local ones
             * unless the DM starts polling the driver after calls
             */
            if ( environment -> error.internal_count > 0 &&
                    rec_number == environment -> error.internal_count )
            {
                char msg[ SQL_MAX_MESSAGE_LENGTH + 40 ];

                if ( rec_number > 1 )
                {
                    sprintf( environment -> msg, 
                            "\n\t\tExit:[%s]",
                                __get_return_status( SQL_NO_DATA ));

                    logPushMsg( environment -> log_handle, 
                            __get_pid( s0 ), 
                            __FILE__, 
                            __LINE__, 
                            LOG_INFO, 
                            LOG_INFO, 
                            environment -> msg );

                    thread_release( SQL_HANDLE_ENV, environment );

                    return SQL_NO_DATA;
                }

                sprintf( msg, "[unixODBC Driver Manager]%s",
                        environment -> error.msg );

                if ( sqlstate )
                    strcpy( sqlstate, environment -> error.sqlstate );

                if ( buffer_length < strlen( msg ))
                {
                    if ( message_text )
                    {
                        memcpy( message_text,
                                msg,
                                buffer_length - 1 );
                        message_text[ buffer_length - 1 ] = '\0';
                    }
                    ret = SQL_SUCCESS_WITH_INFO;
                }
                else
                {
                    if ( message_text )
                        strcpy( message_text, msg );
                     ret = SQL_SUCCESS;
                }
                if ( text_length_ptr )
                    *text_length_ptr = strlen( msg );

                if ( native )
                    *native = environment -> error.native_error;

                sprintf( environment -> msg, 
                        "\n\t\tExit:[%s]\
                        \n\t\t\tSQLState = %s\
                        \n\t\t\tNative = %s\
                        \n\t\t\tMessage Text = %s",
                            __get_return_status( ret ),
                            sqlstate,
                            __ptr_as_string( s1, native ),
                            __sdata_as_string( s1, SQL_CHAR, 
                                text_length_ptr, message_text ));

                logPushMsg( environment -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        environment -> msg );

                thread_release( SQL_HANDLE_ENV, environment );

                return ret;
            }
            else
            {
                sprintf( environment -> msg, 
                        "\n\t\tExit:[%s]",
                            __get_return_status( SQL_NO_DATA ));

                logPushMsg( environment -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        environment -> msg );

                thread_release( SQL_HANDLE_ENV, environment );

                return SQL_NO_DATA;
            }
        }
        break;

      case SQL_HANDLE_DBC:
        {
            DMHDBC connection = (DMHDBC) handle;

            /*
             * check connection
             */

            if ( !__validate_dbc( connection ))
            {
                return SQL_INVALID_HANDLE;
            }

            sprintf( connection -> msg, "\n\t\tEntry:\
                \n\t\t\tConnection = %p\
                \n\t\t\tRec Number = %d\
                \n\t\t\tSQLState = %p\
                \n\t\t\tNative = %p\
                \n\t\t\tMessage Text = %p\
                \n\t\t\tBuffer Length = %d\
                \n\t\t\tText Len Ptr = %p",
                    connection,
                    rec_number,
                    sqlstate,
                    native,
                    message_text,
                    buffer_length,
                    text_length_ptr );

            logPushMsg( connection -> log_handle, 
                    __get_pid( s0 ), 
                    __FILE__, 
                    __LINE__, 
                    LOG_INFO, 
                    LOG_INFO, 
                    connection -> msg );

            thread_protect( SQL_HANDLE_DBC, connection );

            if ( connection -> error.internal_count > 0 &&
                    rec_number == connection -> error.internal_count )
            {
                char msg[ SQL_MAX_MESSAGE_LENGTH + 40 ];

                if ( rec_number > 1 )
                {
                    sprintf( connection -> msg, 
                            "\n\t\tExit:[%s]",
                                __get_return_status( SQL_NO_DATA ));

                    logPushMsg( connection -> log_handle, 
                            __get_pid( s0 ), 
                            __FILE__, 
                            __LINE__, 
                            LOG_INFO, 
                            LOG_INFO, 
                            connection -> msg );

                    thread_release( SQL_HANDLE_DBC, connection );

                    return SQL_NO_DATA;
                }

                sprintf( msg, "[unixODBC Driver Manager]%s",
                        connection -> error.msg );

                if ( sqlstate )
                    strcpy( sqlstate, connection -> error.sqlstate );

                if ( buffer_length < strlen( msg ))
                {
                    if ( message_text )
                    {
                        memcpy( message_text,
                                msg,
                                buffer_length - 1 );
                        message_text[ buffer_length - 1 ] = '\0';
                    }
                    ret = SQL_SUCCESS_WITH_INFO;
                }
                else
                {
                    if ( message_text )
                        strcpy( message_text, msg );
                     ret = SQL_SUCCESS;
                }
                if ( text_length_ptr )
                    *text_length_ptr = strlen( msg );

                if ( native )
                    *native = connection -> error.native_error;

                sprintf( connection -> msg, 
                        "\n\t\tExit:[%s]\
                        \n\t\t\tSQLState = %s\
                        \n\t\t\tNative = %s\
                        \n\t\t\tMessage Text = %s",
                            __get_return_status( ret ),
                            sqlstate,
                            __ptr_as_string( s1, native ),
                            __sdata_as_string( s1, SQL_CHAR, 
                                text_length_ptr, message_text ));

                logPushMsg( connection -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        connection -> msg );

                thread_release( SQL_HANDLE_DBC, connection );

                return ret;
            }
            else if ( connection -> state < STATE_C4 ||
                    !CHECK_SQLGETDIAGREC( connection ))
            {
                sprintf( connection -> msg, 
                        "\n\t\tExit:[%s]",
                            __get_return_status( SQL_NO_DATA ));

                logPushMsg( connection -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        connection -> msg );

                thread_release( SQL_HANDLE_DBC, connection );

                return SQL_NO_DATA;
            }
            else
            {
                rec_number -= connection -> error.internal_count;
                ret = SQLGETDIAGREC( connection,
                                    handle_type,
                                    connection -> driver_dbc,
                                    rec_number,
                                    sqlstate,
                                    native,
                                    message_text,
                                    buffer_length,
                                    text_length_ptr );

                if ( SQL_SUCCEEDED( ret ))
                {
                    sprintf( connection -> msg, 
                        "\n\t\tExit:[%s]\
                        \n\t\t\tSQLState = %s\
                        \n\t\t\tNative = %s\
                        \n\t\t\tMessage Text = %s",
                            __get_return_status( ret ),
                            sqlstate,
                            __ptr_as_string( s1, native ),
                            __sdata_as_string( s1, SQL_CHAR, 
                                text_length_ptr, message_text ));
                }
                else
                {
                    sprintf( connection -> msg, 
                        "\n\t\tExit:[%s]",
                            __get_return_status( ret ));
                }

                logPushMsg( connection -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        connection -> msg );
            }
            thread_release( SQL_HANDLE_DBC, connection );

            return ret;
        }

      case SQL_HANDLE_STMT:
        {
            DMHSTMT statement  = (DMHSTMT) handle;

            /*
             * check statement
             */

            if ( !__validate_stmt( statement ))
            {
                return SQL_INVALID_HANDLE;
            }

            sprintf( statement -> msg, "\n\t\tEntry:\
                \n\t\t\tStatement = %p\
                \n\t\t\tRec Number = %d\
                \n\t\t\tSQLState = %p\
                \n\t\t\tNative = %p\
                \n\t\t\tMessage Text = %p\
                \n\t\t\tBuffer Length = %d\
                \n\t\t\tText Len Ptr = %p",
                    statement,
                    rec_number,
                    sqlstate,
                    native,
                    message_text,
                    buffer_length,
                    text_length_ptr );

            logPushMsg( statement -> log_handle, 
                    __get_pid( s0 ), 
                    __FILE__, 
                    __LINE__, 
                    LOG_INFO, 
                    LOG_INFO, 
                    statement -> msg );

            thread_protect( SQL_HANDLE_STMT, statement );

            if ( statement -> error.internal_count > 0 &&
                    rec_number == statement -> error.internal_count )
            {
                char msg[ SQL_MAX_MESSAGE_LENGTH + 40 ];

                if ( rec_number > 1 )
                {
                    sprintf( statement -> msg, 
                            "\n\t\tExit:[%s]",
                                __get_return_status( SQL_NO_DATA ));

                    logPushMsg( statement -> log_handle, 
                            __get_pid( s0 ), 
                            __FILE__, 
                            __LINE__, 
                            LOG_INFO, 
                            LOG_INFO, 
                            statement -> msg );

                    thread_release( SQL_HANDLE_STMT, statement );

                    return SQL_NO_DATA;
                }

                sprintf( msg, "[unixODBC Driver Manager]%s",
                        statement -> error.msg );

                if ( sqlstate )
                    strcpy( sqlstate, statement -> error.sqlstate );

                if ( buffer_length < strlen( msg ))
                {
                    if ( message_text )
                    {
                        memcpy( message_text,
                                msg,
                                buffer_length - 1 );
                        message_text[ buffer_length - 1 ] = '\0';
                    }
                    ret = SQL_SUCCESS_WITH_INFO;
                }
                else
                {
                    if ( message_text )
                        strcpy( message_text, msg );
                     ret = SQL_SUCCESS;
                }
                if ( text_length_ptr )
                    *text_length_ptr = strlen( msg );

                if ( native )
                    *native = statement -> error.native_error;

                sprintf( statement -> msg, 
                        "\n\t\tExit:[%s]\
                        \n\t\t\tSQLState = %s\
                        \n\t\t\tNative = %s\
                        \n\t\t\tMessage Text = %s",
                            __get_return_status( ret ),
                            sqlstate,
                            __ptr_as_string( s1, native ),
                            __sdata_as_string( s1, SQL_CHAR, 
                                text_length_ptr, message_text ));

                logPushMsg( statement -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        statement -> msg );

                thread_release( SQL_HANDLE_STMT, statement );

                return ret;
            }
            else if ( !CHECK_SQLGETDIAGREC( statement -> connection ))
            {
                sprintf( statement -> msg, 
                        "\n\t\tExit:[%s]",
                            __get_return_status( SQL_NO_DATA ));

                logPushMsg( statement -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        statement -> msg );

                thread_release( SQL_HANDLE_STMT, statement );

                return SQL_NO_DATA;
            }
            else
            {
                rec_number -= statement -> error.internal_count;
                ret = SQLGETDIAGREC( statement -> connection,
                                    handle_type,
                                    statement -> driver_stmt,
                                    rec_number,
                                    sqlstate,
                                    native,
                                    message_text,
                                    buffer_length,
                                    text_length_ptr );

                if ( SQL_SUCCEEDED( ret ))
                {
                    sprintf( statement -> msg, 
                        "\n\t\tExit:[%s]\
                        \n\t\t\tSQLState = %s\
                        \n\t\t\tNative = %s\
                        \n\t\t\tMessage Text = %s",
                            __get_return_status( ret ),
                            sqlstate,
                            __ptr_as_string( s1, native ),
                            __sdata_as_string( s1, SQL_CHAR, 
                                text_length_ptr, message_text ));
                }
                else
                {
                    sprintf( statement -> msg, 
                        "\n\t\tExit:[%s]",
                            __get_return_status( ret ));
                }

                logPushMsg( statement -> log_handle, 
                        __get_pid( s0 ), 
                        __FILE__, 
                        __LINE__, 
                        LOG_INFO, 
                        LOG_INFO, 
                        statement -> msg );
            }

            thread_release( SQL_HANDLE_STMT, statement );

            return ret;
        }
        return SQL_ERROR;

      case SQL_HANDLE_DESC:
        return SQL_ERROR;

      default:
        return SQL_INVALID_HANDLE;
    }
}
