/*********************************************************************
 *
 * This is based on code created by Peter Harvey,
 * (pharvey@codebydesign.com).
 *
 * Modified and extended by Nick Gorham
 * (nick@easysoft.com).
 *
 * Any bugs or problems should be considered the fault of Nick and not
 * Peter.
 *
 * copyright (c) 1999 Nick Gorham
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 **********************************************************************
 *
 * $Id: __connection.c,v 1.4 1999/08/03 21:47:39 shandyb Exp $
 *
 * $Log: __connection.c,v $
 * Revision 1.4  1999/08/03 21:47:39  shandyb
 * Moving to automake: changed files in DriverManager
 *
 * Revision 1.3  1999/07/04 21:05:08  ngorham
 *
 * Add LGPL Headers to code
 *
 * Revision 1.2  1999/06/19 17:51:41  ngorham
 *
 * Applied assorted minor bug fixes
 *
 * Revision 1.1.1.1  1999/05/29 13:41:09  sShandyb
 * first go at it
 *
 * Revision 1.1.1.1  1999/05/27 18:23:18  pharvey
 * Imported sources
 *
 * Revision 1.1  1999/04/25 23:06:11  nick
 * Initial revision
 *
 *
 **********************************************************************/

#include "drivermanager.h"

/*
 * list of places to look, a $ at the start indicates
 * then it following text should be looked in as a env
 * variable
 */

char *ini_search_list[] =
{
    INIUSER,
    "$ODBCINI",
    INISYSTEM,
};

static char const rcsid[]= "$RCSfile: __connection.c,v $ $Revision: 1.4 $";

/*
 * search for the library (.so) that the DSN points to
 */

char *__find_lib_name( char *dsn, char *lib_name, char *driver_name )
{
    char driver[ INI_MAX_PROPERTY_VALUE + 1 ];
    char driver_lib[ INI_MAX_PROPERTY_VALUE + 1 ];

    SQLSetConfigMode( ODBC_USER_DSN );

	/*
	 * GET DRIVER FROM ODBC.INI
	 */

    SQLGetPrivateProfileString( dsn, "Driver", "",
            driver_lib, sizeof( driver_lib ), "ODBC.INI" );

    if ( driver_lib[ 0 ] == 0 )
    {
        /*
         * if not found look in system DSN
         */

        SQLSetConfigMode( ODBC_SYSTEM_DSN );

        SQLGetPrivateProfileString( dsn, "Driver", "",
                driver_lib, sizeof( driver_lib ), "ODBC.INI" );
        
        SQLSetConfigMode( ODBC_BOTH_DSN );
        if ( driver_lib[ 0 ] == 0 )
            return NULL;
    }

	/*
	 * GET DRIVER FROM ODBCINST.INI IF ODBC.INI HAD USER FRIENDLY NAME
	 */

    strcpy( driver_name, "" );

    if ( driver_lib[ 0 ] != '/' )
	{
        strcpy( driver, driver_lib );

		SQLGetPrivateProfileString( driver, "Driver", "",
				driver_lib, sizeof( driver_lib ), "ODBCINST.INI" );

        strcpy( driver_name, driver );

		if ( driver_lib[ 0 ] == 0 )
			return NULL;
	}

	strcpy( lib_name, driver_lib );

    return lib_name;
}

void __set_trace( DMHDBC connection, char *dsn )
{
    char    	szTRACE[INI_MAX_PROPERTY_VALUE+1]		= "";
    char    	szTRACEFILE[INI_MAX_PROPERTY_VALUE+1]	= "";

	SQLGetPrivateProfileString( dsn, "TRACE", "", szTRACE, sizeof(szTRACE), ".odbc.ini" );
	if ( szTRACE )
	{
		if ( szTRACE[0] == '1' || toupper(szTRACE[0]) == 'Y' || toupper(szTRACE[0]) == 'O' )
		{
			SQLGetPrivateProfileString( dsn, "TRACEFILE", "", szTRACEFILE, sizeof(szTRACEFILE), ".odbc.ini" );
			if ( szTRACEFILE )
			{
				if ( connection->log_handle->pszLogFile )
					free( connection->log_handle->pszLogFile );
                connection->log_handle->pszLogFile = NULL;
				
				connection->log_handle->pszLogFile = (char*)strdup( szTRACEFILE );
			}
		}
	}
}



